import 'dart:convert';
import 'package:crypto/crypto.dart';
// ignore: implementation_imports
import 'package:crypto/src/digest_sink.dart';

import 'device.dart';

abstract class IGAPIConfig {

  static const NAME = "Instagram";
  static const VERSION = "97.0.0.11.120";
  static const BUILD = "157869373";
  static const CAPABILITIES = "36r/dw==";
  static const FACEBOOK_ID = "124024574287414";
  static const LOG_SECRET_KEY = "84a456d620314b6e92a16d8ff1c792dc";
  static const SECRET_KEY_1 =
      "BQJXVFUADwUDUwIOAA5SBVMBUgQHUwAPAwQBUgYPAQ9QVVIFAAIFBVRSDgEGUFVVVAdSBwMGUw4FVwUPDlMBVA==";
  static const SECRET_KEY_2 =
      "b2g9Pj9qZW9pOWhkamQ4bzlrOG5tOWplaW5rOGxla2U6Pzhvamhvbz44ZGtsOj8/Pm04bWlsOWRvPW9lZDlrPg==";

  static const API_ORIGIN = "https://i.instagram.com";
  static const API_V1 = API_ORIGIN + "/api/v1";
  static const API_V2 = API_ORIGIN + "/api/v2";

  static String getUserAgent(QuarkDevice device) {
    final version = device.firmwareVersion.replaceAll(".", "_");
    return "$NAME $VERSION (${device.deviceType}; iOS $version; en_US; en-US; scale=${device.display.nativeScale.toStringAsFixed(2)}; gamut=normal; ${device.display.resolutionToString(reverse: true)}; $BUILD) AppleWebKit/420+";
  }

  static get commonHeaders {
    return {
      "X-IG-Connection-Type": "WiFi",
      "X-IG-Capabilities": CAPABILITIES,
      "X-IG-App-ID": FACEBOOK_ID,
      "X-IG-ABR-Connection-Speed-KBPS": "0",
      "X-IG-Connection-Speed": "-1kbps",
      "X-IG-Bandwidth-Speed-KBPS": "0.000",
    };
  }

  static Map<String, String> signatureBody(Map<String, dynamic> json) {
    var jsonstr = jsonEncode(json);
    var signedBody = signature(jsonstr) + "." + jsonstr;
    return {"signed_body": signedBody, "ig_sig_key_version": "5"};
  }

  static String signature(String string) {
    var body = utf8.encode(string);
    var key1 = base64Decode(SECRET_KEY_1);
    var key2 = base64Decode(SECRET_KEY_2);

    var ds = new DigestSink();
    var s = sha256.startChunkedConversion(ds);
    s.add(key1);
    s.add(body);
    s.close();

    var ds2 = new DigestSink();
    var s2 = sha256.startChunkedConversion(ds2);
    s2.add(key2);
    s2.add(ds.value.bytes);
    s2.close();

    return ds2.value.toString();
  }
}
