import 'package:igapi/igapi.dart';
import 'package:igapi/src/actions.dart';

/// Makers 不公开

enum InstructionType { Terminator, Push, Pop, Cast }

abstract class Instruction {
  InstructionType get type;
}

class _Cvrt {
  static bool isNull(dynamic val) => val == null;

  static bool isEmpty(dynamic val) => getCount(val) == 0;

  static T? as<T>(dynamic val) => val is T ? val : null;

  static int getCount(dynamic val) {
    if (val is Map) return (val).length;
    if (val is List) return (val).length;
    if (val is String) return (val).length;
    return 0;
  }

  static dynamic valueForKey(dynamic val, String key) {
    if (val is Map) return (val)[key];
    return null;
  }

  static dynamic firstElementOrSelf(dynamic val) {
    if (val is List)
      return val.first;
    else
      return val;
  }
}

enum TerminatorCondType {
  Unconditional,
  NoKey,
  NotBoolean,
  EqualToBoolean,
  NotNumber,
  EqualToNumber,
  NotEqualToNumber,
  NotString,
  EqualToString,
  NotEqualToString,
  NotMap,
  NotList,
  Empty
}

class TerminatorInst extends Instruction {
  TerminatorCondType _condType;

  String? _key;
  bool _checkTarget;

  bool? _throws;
  bool? _bitI;
  int? _numI;
  String? _strI;

  TerminatorInst()
      : _condType = TerminatorCondType.Unconditional,
        _key = null,
        _checkTarget = false;

  @override
  InstructionType get type => InstructionType.Terminator;

  TerminatorInst get source {
    _checkTarget = false;
    return this;
  }

  TerminatorInst get target {
    _checkTarget = true;
    return this;
  }

  TerminatorInst get uncond {
    _condType = TerminatorCondType.Unconditional;
    _key = null;
    return this;
  }

  TerminatorInst nkey(String key) {
    _condType = TerminatorCondType.NoKey;
    _key = key;
    return this;
  }

  TerminatorInst nbit(String key) {
    _condType = TerminatorCondType.NotBoolean;
    _key = key;
    return this;
  }

  TerminatorInst eqbit(String key, bool bit) {
    _condType = TerminatorCondType.EqualToBoolean;
    _key = key;
    _bitI = bit;
    return this;
  }

  TerminatorInst nnum(String key) {
    _condType = TerminatorCondType.NotNumber;
    _key = key;
    return this;
  }

  TerminatorInst eqnum(String key, int number) {
    _condType = TerminatorCondType.EqualToNumber;
    _key = key;
    _numI = number;
    return this;
  }

  TerminatorInst neqnum(String key, int number) {
    _condType = TerminatorCondType.NotEqualToNumber;
    _key = key;
    _numI = number;
    return this;
  }

  TerminatorInst nstr(String key) {
    _condType = TerminatorCondType.NotString;
    _key = key;
    return this;
  }

  TerminatorInst eqstr(String key, String string) {
    _condType = TerminatorCondType.EqualToString;
    _key = key;
    _strI = string;
    return this;
  }

  TerminatorInst neqstr(String key, String string) {
    _condType = TerminatorCondType.NotEqualToString;
    _key = key;
    _strI = string;
    return this;
  }

  TerminatorInst nobj(String key) {
    _condType = TerminatorCondType.NotMap;
    _key = key;
    return this;
  }

  TerminatorInst narr(String key) {
    _condType = TerminatorCondType.NotList;
    _key = key;
    return this;
  }

  TerminatorInst empty(String key) {
    _condType = TerminatorCondType.Empty;
    _key = key;
    return this;
  }

  TerminatorInst get throws {
    _throws = true;
    return this;
  }

  String? get errorMessage {
    if (_throws != true) return null;

    switch (_condType) {
      case TerminatorCondType.Unconditional:
        return null;
      case TerminatorCondType.NoKey:
        return "val:$_key does not exist.";
      case TerminatorCondType.NotBoolean:
        return "val:$_key is not bool.";
      case TerminatorCondType.EqualToBoolean:
        return "val:$_key should not be $_bitI.";
      case TerminatorCondType.NotNumber:
        return "val:$_key is not number.";
      case TerminatorCondType.EqualToNumber:
        return "val:$_key should not be equal to $_numI.";
      case TerminatorCondType.NotEqualToNumber:
        return "val:$_key should be equal to $_numI.";
      case TerminatorCondType.NotString:
        return "val:$_key is not string.";
      case TerminatorCondType.EqualToString:
        return "val:$_key should not be equal to $_strI.";
      case TerminatorCondType.NotEqualToString:
        return "val:$_key should be equal to $_strI.";
      case TerminatorCondType.NotMap:
        return "val:$_key is not map.";
      case TerminatorCondType.NotList:
        return "val:$_key is not list.";
      case TerminatorCondType.Empty:
        return "val:$_key should not be empty.";
    }
  }

  bool exec(Map<String, dynamic>? target, Map<String, dynamic>? source) {
    Map<String, dynamic>? container = _checkTarget == true ? target : source;
    if (container == null) return true;

    final val = container[_key];

    switch (_condType) {
      case TerminatorCondType.Unconditional:
        return true;
      case TerminatorCondType.NoKey:
        return _Cvrt.isNull(val);
      case TerminatorCondType.NotBoolean:
        return _Cvrt.as<bool>(val) == null;
      case TerminatorCondType.EqualToBoolean:
        return _bitI == _Cvrt.as<bool>(val);
      case TerminatorCondType.NotNumber:
        return _Cvrt.as<num>(val) == null;
      case TerminatorCondType.EqualToNumber:
        return _numI == _Cvrt.as<num>(val);
      case TerminatorCondType.NotEqualToNumber:
        return _numI != _Cvrt.as<num>(val);
      case TerminatorCondType.NotString:
        return _Cvrt.as<String>(val) == null;
      case TerminatorCondType.EqualToString:
        return _strI == _Cvrt.as<String>(val);
      case TerminatorCondType.NotEqualToString:
        return _strI != _Cvrt.as<String>(val);
      case TerminatorCondType.NotMap:
        return _Cvrt.as<Map>(val) == null;
      case TerminatorCondType.NotList:
        return _Cvrt.as<List>(val) == null;
      case TerminatorCondType.Empty:
        return _Cvrt.isEmpty(val);
    }
  }
}

class PushInst extends Instruction {
  final String key;

  @override
  InstructionType get type => InstructionType.Push;

  PushInst(this.key);
}

class PopInst extends Instruction {
  @override
  InstructionType get type => InstructionType.Pop;
}

enum CastOutputType { One, List, ListOne, List2Map, Map }

abstract class CastInst <T> extends Instruction {
  final String target;
  final String source;

  CastOutputType _outputType = CastOutputType.One;
  List<String>? _infilterKeyPath;
  List<String>? _extractKeyPath;
  String? _mapkey;
  TerminatorInst? _skip;

  @override
  InstructionType get type => InstructionType.Cast;

  CastInst(String target, String? source)
      : target = target,
        source = source ?? target;

  TerminatorInst get skip {
    _skip = TerminatorInst();
    return _skip!;
  }

  CastInst get one {
    _outputType = CastOutputType.One;
    return this;
  }

  CastInst get list {
    _outputType = CastOutputType.List;
    return this;
  }

  CastInst get listone {
    _outputType = CastOutputType.ListOne;
    return this;
  }

  CastInst list2map(String key) {
    _outputType = CastOutputType.List2Map;
    _mapkey = key;
    return this;
  }

  CastInst get map {
    _outputType = CastOutputType.Map;
    return this;
  }

  CastInst infilter(List<String> keyPath) {
    _infilterKeyPath = keyPath;
    return this;
  }

  CastInst extract(List<String> keyPath) {
    _extractKeyPath = keyPath;
    return this;
  }

  dynamic cast(dynamic val);

  List get listO {
    List<T> v = [];
    return v;
  }
  Map get mapO {
    Map<String, T> v = {};
    return v;
  }

  void exec(Map<String, dynamic> targetC, Map<String, dynamic>? sourceC) {
    sourceC = sourceC ?? {};
    bool shouldSkip = _skip != null && _skip!.exec(targetC, sourceC);
    if (shouldSkip) return;

    var ret;
    var src;
    if (source == "@this")
      src = sourceC;
    else
      src = sourceC[source];

    _infilterKeyPath?.forEach((v) {
      src = _Cvrt.valueForKey(src, v);
    });

    switch (_outputType) {
      case CastOutputType.One:
        var one = _Cvrt.firstElementOrSelf(src);
        _extractKeyPath?.forEach((v) {
          one = _Cvrt.valueForKey(one, v);
        });
        ret = cast(one);
        break;
      case CastOutputType.List:
      case CastOutputType.ListOne:
      case CastOutputType.List2Map:
        List list;
        if (src is Map) list = [src];
        else if (src is List) list = src;
        else list = [];
        List targets = listO;
        if (src == null)
          list = [];
        else if (src is! List) list = [src];
        list.forEach((v) {
          var one = v;
          _extractKeyPath?.forEach((v) {
            one = _Cvrt.valueForKey(one, v);
          });
          var target = cast(one);
          if (target != null) targets.add(target);
        });

        if (_outputType == CastOutputType.ListOne) {
          ret = targets.length > 0 ? targets.first : null;
        } else if (_outputType == CastOutputType.List2Map) {
          Map map = mapO;
          targets.forEach((v) {
            if (v is Map) {
              final key = v[_mapkey];
              if (key is String) {
                map[key] = v;
              }
            }
          });
          ret = map;
        } else {
          ret = targets;
        }
        break;
      case CastOutputType.Map:
        Map targets = mapO;
        if (src is Map) {
          (src as Map).forEach((k, v) {
            var one = v;
            _extractKeyPath?.forEach((v) {
              one = _Cvrt.valueForKey(one, v);
            });
            var target = cast(one);
            if (target != null) targets[k] = target;
          });
        }
        ret = targets;
        break;
    }
    targetC[target] = ret;
  }
}

class BooleanCastInst extends CastInst<bool> {
  int _format = 0;
  String? _key;
  late bool _ret;
  bool _unkas = false;

  BooleanCastInst(String target, String? source) : super(target, source);

  BooleanCastInst isstr(bool ret) {
    _format = 1;
    _ret = ret;
    return this;
  }

  BooleanCastInst eqstr(String str, bool ret) {
    _format = 2;
    _key = str;
    _ret = ret;
    return this;
  }

  BooleanCastInst unkas(bool unkas) {
    _unkas = unkas;
    return this;
  }

  dynamic cast(dynamic val) {
    if (val is Map || val is List) val = null;

    if (_Cvrt.isNull(val)) return _unkas;
    if (val is bool) return val;
    if (val is num) return val != 0;
    if (val is String) {
      if (_format == 0)
        return _unkas;
      else if (_format == 1)
        return _ret;
      else if (_format == 2) {
        if (val == _key) {
          return _ret ? true : false;
        } else {
          return _ret ? false : true;
        }
      }
    }
    return _unkas;
  }
}

class NumberCastInst extends CastInst<num> {
  bool _asInt = false;

  NumberCastInst(String target, String? source) : super(target, source);

  NumberCastInst get inum {
    _asInt = true;
    return this;
  }

  NumberCastInst get fnum {
    _asInt = false;
    return this;
  }

  dynamic cast(dynamic val) {
    if (val == null || val is Map || val is List) return 0;
    if (val is bool) return val ? 1 : 0;
    if (val is num) {
      if (_asInt == true) return val.toInt();
      return val;
    }
    if (val is String) {
      if (_asInt == true) return int.tryParse(val) ?? 0;
      return num.tryParse(val) ?? 0;
    }
    return 0;
  }
}

class EnumCastInst extends CastInst<int> {
  late Map<String, int> _map;

  EnumCastInst(String target, String? source) : super(target, source);

  EnumCastInst strmap(Map<String, int> map) {
    _map = map;
    return this;
  }

  dynamic cast(dynamic val) {
    var senum = _map[val.toString()];
    if (senum == null) senum = _map["@uncaught"] ?? 0;
    return senum;
  }
}

class StringCastInst extends CastInst<String> {
  StringCastInst(String target, String? source) : super(target, source);

  dynamic cast(dynamic val) {
    if (val == null || val is Map || val is List) return null;
    return val.toString();
  }
}

class URLCastInst extends CastInst<String> {
  String? _base;
  late bool _directory;

  URLCastInst(String target, String? source) : super(target, source);

  URLCastInst base(String? baseUrl, [bool? directory]) {
    _base = baseUrl;
    _directory = directory ?? false;
    return this;
  }

  dynamic cast(dynamic val) {
    if (val is! String) return null;
    if (_base != null) {
      return _base! + val + (_directory ? "/" : "");
    }
    if (val.startsWith(RegExp("(https?|ftp|file)://"))) return val;
    return null;
  }
}

class DateCastInst extends CastInst<int> {
  bool _asTS10IfNum = false;

  DateCastInst(String target, String? source) : super(target, source);

  DateCastInst get ts10 {
    _asTS10IfNum = true;
    return this;
  }

  DateCastInst get ts13 {
    _asTS10IfNum = false;
    return this;
  }

  dynamic cast(dynamic val) {
    if (val is num) {
      return _asTS10IfNum ? val * 1000 : val;
    } else if (val is String) {
      return DateTime.tryParse(val)?.millisecondsSinceEpoch ?? 0;
    } else
      return 0;
  }
}

class PointCastInst extends CastInst<SXPoint> {
  String _x = "x";
  String _y = "y";

  PointCastInst(String target, String? source) : super(target, source);

  PointCastInst x(String key) {
    _x = key;
    return this;
  }

  PointCastInst y(String key) {
    _y = key;
    return this;
  }

  dynamic cast(dynamic val) {
    if (val is! Map && val is! List) return null;

    num? x = 0.0;
    num? y = 0.0;
    if (val is List) {
      var ix = int.tryParse(_x);
      var iy = int.tryParse(_y);
      if (ix == null || iy == null) return null;
      if (ix >= val.length || iy >= val.length) return null;
      x = val[ix];
      y = val[iy];
    } else {
      x = val[_x];
      y = val[_y];
    }

    if (x is! num || y is! num) return null;
    return QuarkModel.fromJson<SXPoint>({"x": x, "y": y});
  }
}

class SizeCastInst extends CastInst<SXSize> {
  String _width = "width";
  String _height = "height";

  SizeCastInst(String target, String? source) : super(target, source);

  SizeCastInst width(String key) {
    _width = key;
    return this;
  }

  SizeCastInst height(String key) {
    _height = key;
    return this;
  }

  dynamic cast(dynamic val) {
    if (val is! Map && val is! List) return null;

    num? x = 0.0;
    num? y = 0.0;
    if (val is List) {
      var ix = int.tryParse(_width);
      var iy = int.tryParse(_height);
      if (ix == null || iy == null) return null;
      if (ix >= val.length || iy >= val.length) return null;
      x = val[ix];
      y = val[iy];
    } else {
      x = val[_width];
      y = val[_height];
    }

    if (x is! num || y is! num) return null;
    return QuarkModel.fromJson<SXSize>({"width": x, "height": y});
  }
}

typedef Map MapGenFunc();
typedef List ListGenFunc();

class ModelCastInst extends CastInst<QuarkModel> {
  late Type _cls;
  late MapGenFunc _mapGen;
  late ListGenFunc _listGen;
  String _tag = "default";

  ModelCastInst(String target, String? source) : super(target, source);

  ModelCastInst cls<T extends QuarkModel>([String tag = "default"]) {
    _cls = T;
    _tag = tag;
    
    _mapGen = () {
      Map<String, T> map = {};
      return map;
    };
    _listGen = () {
      List<T> list = [];
      return list;
    };
    return this;
  }

  List get listO => _listGen();
  Map get mapO => _mapGen();

  dynamic cast(dynamic val) {
    if (val is! Map<String, dynamic>) return null;
    return MM.create(_cls, val, _tag);
  }
}

class ModelMaker {
  List<Instruction> insts = [];

  TerminatorInst get retnil {
    final inst = TerminatorInst();
    insts.add(inst);
    return inst;
  }

  PushInst push(String key) {
    final inst = PushInst(key);
    insts.add(inst);
    return inst;
  }

  PopInst get pop {
    final inst = PopInst();
    insts.add(inst);
    return inst;
  }

  BooleanCastInst bit(String target, [String? source]) {
    final inst = BooleanCastInst(target, source);
    insts.add(inst);
    return inst;
  }

  NumberCastInst num(String target, [String? source]) {
    final inst = NumberCastInst(target, source);
    insts.add(inst);
    return inst;
  }

  EnumCastInst senum(String target, [String? source]) {
    final inst = EnumCastInst(target, source);
    insts.add(inst);
    return inst;
  }

  StringCastInst str(String target, [String? source]) {
    final inst = StringCastInst(target, source);
    insts.add(inst);
    return inst;
  }

  URLCastInst url(String target, [String? source]) {
    final inst = URLCastInst(target, source);
    insts.add(inst);
    return inst;
  }

  DateCastInst date(String target, [String? source]) {
    final inst = DateCastInst(target, source);
    insts.add(inst);
    return inst;
  }

  PointCastInst point(String target, [String? source]) {
    final inst = PointCastInst(target, source);
    insts.add(inst);
    return inst;
  }

  SizeCastInst size(String target, [String? source]) {
    final inst = SizeCastInst(target, source);
    insts.add(inst);
    return inst;
  }

  ModelCastInst model(String target, [String? source]) {
    final inst = ModelCastInst(target, source);
    insts.add(inst);
    return inst;
  }

  Map<String, dynamic>? parse(Map<String, dynamic>? obj) {
    if (obj is! Map) return null;
    List stack = [obj];
    Map<String, dynamic> target = {};

    for (int i = 0; i < insts.length; i++) {
      final inst = insts.elementAt(i);
      switch (inst.type) {
        case InstructionType.Terminator:
          final term = inst as TerminatorInst;
          bool exit = term.exec(target, stack.last);
          if (exit == true) {
            if (term._throws == true) {
              throw IGException(
                  IGExceptionType.InvalidModel, term.errorMessage ?? "");
            }
            return null;
          }
          break;
        case InstructionType.Push:
          final push = inst as PushInst;
          Map? last = stack.last;
          if (last is Map) {
            stack.add(last[push.key]);
          } else {
            stack.add(null);
          }
          break;
        case InstructionType.Pop:
          if (stack.length > 0) stack.removeLast();
          break;
        case InstructionType.Cast:
          final cast = inst as CastInst;
          cast.exec(target, stack.last);
          break;
      }
    }

    return target;
  }
}

typedef QuarkModel MakeModel();
typedef void MakeMaker(ModelMaker make);

const String VERIFY_MODEL_JSON_TAG = "verify_model_json";

class MM {
  final Type cls;
  final MakeModel gen;
  final Map<String, ModelMaker> makes = {};

  static final Map<Type, MM> _cache = {};

  MM(this.gen) : cls = gen().runtimeType {
    _cache[cls] = this;
  }

  static String _getActionName(IGAPIAction action) => "IGEP#${action.index}";

  MM def(MakeMaker make) {
    return add("default", make);
  }

  MM verify(MakeMaker make) {
    return add(VERIFY_MODEL_JSON_TAG, make);
  }

  MM act(IGAPIAction action, MakeMaker make) {
    return add(_getActionName(action), make);
  }

  MM add(String tag, MakeMaker make) {
    assert(makes[tag] == null);

    ModelMaker maker = ModelMaker();
    make(maker);
    makes[tag] = maker;
    return this;
  }

  static MM endp = MM(() => SXEndpointModel()).def((make) {});

  static T? create<T extends QuarkModel>(Type cls, Map<String, dynamic> obj,
      [String tag = "default"]) {
    modelMakerSetup();
    MM? mm = _cache[cls];
    assert(mm != null);
    final maker = mm!.makes[tag];
    assert(maker != null);
    if (tag == VERIFY_MODEL_JSON_TAG && maker == null) {
      throw IGException(IGExceptionType.InvalidModel, "$cls does not support fromJson.");
    }

    Map<String, dynamic>? target = maker!.parse(obj);
    if (target == null) return null;
    QuarkModel newQuark = mm.gen();
    newQuark.internalAsSelf(target);
    return newQuark as T;
  }

  static T? createFromJson<T extends QuarkModel>(Type cls, Map<String, dynamic> obj) => create(cls, obj, VERIFY_MODEL_JSON_TAG);

  static SXEndpointModel? createEP(IGAPIAction action, Map<String, dynamic> obj) {
    return create(SXEndpointModel, obj, _getActionName(action)) as SXEndpointModel;
  }
}

bool _modelMakerSetupOnceToken = false;
void modelMakerSetup() {
  if (_modelMakerSetupOnceToken == true) return;
  _modelMakerSetupOnceToken = true;

  MM(()=> SXPoint()).verify((make) {
    make.retnil.nkey("x").throws;
    make.retnil.nkey("y").throws;
    make.num("x");
    make.num("y");
  });

  MM(()=> SXSize()).verify((make) {
    make.retnil.nkey("width").throws;
    make.retnil.nkey("height").throws;
    make.num("width");
    make.num("height");
  });

  MM(() => IGPagination()).def((make) {
    make.str('maxId', 'next_max_id');
  }).add("children_comment", (make) {
    make.str('maxId', 'next_max_child_cursor');
  }).add("recommended_user", (make) {
    make.retnil.eqbit('more_available', false); // ig's bug
    make.str('maxId', 'max_id');
  }).add("location_feed", (make) {
    make.str('maxId', 'next_max_id');
    make.num('page', 'next_page').inum;
  }).verify((make) {
    make.str("maxId");
  });

  MM(() => IGTwoFactorLock()).def((make) {
    make.retnil.nkey("two_factor_info").throws;
    make.push("two_factor_info");
    make.retnil.nkey("two_factor_identifier").throws;
    make.str('tfaId', 'two_factor_identifier');
    make.str('username');
    make.str('contact', 'obfuscated_phone_number');
  });

  MM(() => IGChallengeLock()).def((make) {
    make.retnil.nobj("challenge").throws;
    make.push("challenge");
    make.retnil.nkey("api_path").throws;

    make.url('url');
    make.str('apiPath', 'api_path');
  });

  MM(() => IGChallengeSelection()).def((make) {
    make.senum('step', 'step_name').strmap({
      '@uncaught': IGChallengeStep.Uncaught.index,
      'select_verify_method': IGChallengeStep.SelectVerifyMethod.index,
      'delta_login_review': IGChallengeStep.DeltaLoginReview.index,
    });

    make
        .senum('step', 'action')
        .strmap({
          '@uncaught': IGChallengeStep.Uncaught.index,
          'close': IGChallengeStep.Completed.index
        })
        .skip
        .neqstr('action', 'close');

    make.push('step_data');
    make.str('phone', 'phone_number');
    make.str('email');
  });

  MM(() => IGChallengeVerification()).def((make) {
    make.senum('step', 'step_name').strmap({
      '@uncaught': IGChallengeStep.Uncaught.index,
      'verify_email': IGChallengeStep.VerifyByEmail.index,
      'verify_code': IGChallengeStep.VerifyByPhone.index,
    });

    make
        .senum('step', 'action')
        .strmap({
          '@uncaught': IGChallengeStep.Uncaught.index,
          'close': IGChallengeStep.Completed.index
        })
        .skip
        .neqstr('action', 'close');

    make.str("userId", "user_id");
    make.str("nonce", "nonce_code");

    make.push('step_data');
    make.senum('choice', 'form_type').strmap({
      '@uncaught': IGChallengeChoice.Uncaught.index,
      'phone_number': IGChallengeChoice.Phone.index,
      'email': IGChallengeChoice.Email.index,
    });
    make.num('resendDelay', 'resend_delay').fnum;
    make.str('contact', 'contact_point');
  });

  MM(() => IGLoggedInUser()).def((make) {
    make.retnil.nkey("pk").throws;

    make.str('userId', 'pk');
    make.str('username', 'username');
    make.str('nickname', 'full_name');

    make.url('thumbnailAvatarURL', 'profile_pic_url');
    make.bit('anonymousAvatar', 'profile_pic_id').isstr(false).unkas(true);
    make
        .bit('anonymousAvatar', 'has_anonymous_profile_picture')
        .skip
        .nkey('has_anonymous_profile_picture');

    make.bit('private', 'is_private');
    make.bit('verified', 'is_verified');
    make.bit("bussiness", "is_bussiness");

    make.senum("commenterControl", "allowed_commenter_type").strmap({
      "@uncaught": IGCommenterControl.Uncaught.index,
      "any": IGCommenterControl.Any.index,
      "following_and_follower": IGCommenterControl.FollowingAndFollower.index,
      "following": IGCommenterControl.Following.index,
      "follower": IGCommenterControl.Follower.index
    });

    make.str('phone', 'phone_number');
    make.str("fbuid");
  }).verify((make) {
    make.retnil.nkey("userId").throws;

    make.str("userId");
    make.str("username");
    make.str("nickname");
    make.url("thumbnailAvatarURL", "profile_pic_url");
    make.bit("anonymousAvatar").isstr(false).unkas(true);
    make.bit("private");
    make.bit("verified");
    make.bit("bussiness");

    make.num("commenterControl").inum;

    make.str("phone");
    make.str("fbuid");
  });

  MM(() => IGAccountUser()).def((make) {
    make.retnil.nkey("pk").throws;

    make.str('userId', 'pk');
    make.str('username', 'username');
    make.str('nickname', 'full_name');

    make.url('thumbnailAvatarURL', 'profile_pic_url');
    make.bit('anonymousAvatar', 'profile_pic_id').isstr(false).unkas(true);
    make
        .bit('anonymousAvatar', 'has_anonymous_profile_picture')
        .skip
        .nkey('has_anonymous_profile_picture');
    make
        .model('narrowedAvatarURLs', 'hd_profile_pic_versions')
        .cls<IGMediaResource>()
        .list;
    make
        .model('originalAvatarURL', 'hd_profile_pic_url_info')
        .cls<IGMediaResource>();

    make.bit('private', 'is_private');
    make.bit('verified', 'is_verified');

    make.str('biography');
    make.url('website', 'external_url');

    make.senum("commenterControl", "allowed_commenter_type").strmap({
      "@uncaught": IGCommenterControl.Uncaught.index,
      "any": IGCommenterControl.Any.index,
      "following_and_follower": IGCommenterControl.FollowingAndFollower.index,
      "following": IGCommenterControl.Following.index,
      "follower": IGCommenterControl.Follower.index
    });

    make.str('email');
    make.str('phone', 'phone_number');

    make.senum('gender').strmap({
      '@uncaught': IGUserGender.Uncaught.index,
      '1': IGUserGender.Male.index,
      '2': IGUserGender.Female.index,
      '3': IGUserGender.NotSpecified.index
    });
  }).verify((make) {
    make.retnil.nkey("userId").throws;

    make.str("userId");
    make.str("username");
    make.str("nickname");
    make.url("thumbnailAvatarURL", "profile_pic_url");
    make.bit("anonymousAvatar").isstr(false).unkas(true);

    make.model("narrowedAvatarURLs")
        .cls<IGMediaResource>(VERIFY_MODEL_JSON_TAG).list;
    make.model("originalAvatarURL").cls<IGMediaResource>(VERIFY_MODEL_JSON_TAG);

    make.bit("private");
    make.bit("verified");

    make.str("biography");
    make.url("website");

    make.num("commenterControl").inum;

    make.str("email");
    make.str("phone");

    make.num("gender").inum;
  });

  MM(() => IGUser()).def((make) {
    make.retnil.nkey("pk").throws;

    make.str('userId', 'pk');
    make.str('username', 'username');
    make.str('nickname', 'full_name');

    make.url('thumbnailAvatarURL', 'profile_pic_url');
    make.bit('anonymousAvatar', 'profile_pic_id').isstr(false).unkas(true);
    make
        .bit('anonymousAvatar', 'has_anonymous_profile_picture')
        .skip
        .nkey('has_anonymous_profile_picture');
    make
        .model('narrowedAvatarURLs', 'hd_profile_pic_versions')
        .cls<IGMediaResource>()
        .list;
    make
        .model('originalAvatarURL', 'hd_profile_pic_url_info')
        .cls<IGMediaResource>();

    make.bit('private', 'is_private');
    make.bit('verified', 'is_verified');
    make.bit('business', 'is_business');
    make.bit('favorite', 'is_favorite');
    make.bit('hasChaining', 'has_chaining');

    make.num('mediaCount', 'media_count').inum;
    make.num('usertagCount', 'usertags_count').inum;
    make.num('followerCount', 'follower_count').inum;
    make.num('followeeCount', 'following_count').inum;

    make.str('biography', 'biography');
    make.url('website', 'external_url');
  }).verify((make) {
    make.retnil.nkey("userId").throws;

    make.str("userId");
    make.str("username");
    make.str("nickname");
    make.url("thumbnailAvatarURL");
    make.bit("anonymousAvatar").isstr(false).unkas(true);

    make.model("narrowedAvatarURLs")
        .cls<IGMediaResource>(VERIFY_MODEL_JSON_TAG).list;
    make.model("originalAvatarURL").cls<IGMediaResource>(VERIFY_MODEL_JSON_TAG);

    make.bit("private");
    make.bit("verified");

    make.bit("business");
    make.bit("favorite");
    make.bit("hasChaining");

    make.num("mediaCount").inum;
    make.num("usertagCount").inum;
    make.num("followerCount").inum;
    make.num("followeeCount").inum;

    make.str("biography");
    make.url("website");
  });

  MM(() => IGUserLite()).def((make) {
    make.retnil.nkey("pk").throws;

    make.str('userId', 'pk');
    make.str('username', 'username');
    make.str('nickname', 'full_name');

    make.url('thumbnailAvatarURL', 'profile_pic_url');
    make.bit('anonymousAvatar', 'profile_pic_id').isstr(false).unkas(true);
    make
        .bit('anonymousAvatar', 'has_anonymous_profile_picture')
        .skip
        .nkey('has_anonymous_profile_picture');
    make.bit('private', 'is_private');
    make.bit('verified', 'is_verified');
  }).verify((make) {
    make.retnil.nkey("userId").throws;
    
    make.str("userId");
    make.str("username");
    make.str("nickname");

    make.url("thumbnailAvatarURL");
    make.bit("anonymousAvatar").isstr(false).unkas(true);
    make.bit("private");
    make.bit("verified");
  });

  MM(() => IGMediaResource()).def((make) {
    make.url('url', 'url');
    make.size('size', '@this').width('width').height('height');
    make.str('representationId', 'id').skip.nkey('id');
  }).verify((make) {
    make.url("url");
    make.size("size");
    make.str("representationId");
  });

  MM(() => IGUserGraphStatus()).def((make) {
    make.bit('following');
    make.bit('private', 'is_private');
    make.bit('incoming', 'incoming_request');
    make.bit('outgoing', 'outgoing_request');
    make.bit('bestie', 'is_bestie');
    make.bit('blocking');
    make.bit('followedBy', 'followed_by');
  }).verify((make) {
    make.bit("following");
    make.bit("private");
    make.bit("incoming");
    make.bit("outgoing");
    make.bit("bestie");
    make.bit("blocking");
    make.bit("followedBy");
  });

  MM(() => IGUsertag()).def((make) {
    make.model('user', 'user').cls<IGUserLite>();
    make.point('position', 'position').x('0').y('1');
  }).verify((make) {
    make.model("user").cls<IGUserLite>(VERIFY_MODEL_JSON_TAG);
    make.point("position");
  });

  MM(() => IGMediaItem()).def((make) {
    make.retnil.nkey("id").throws;

    make.str('mediaId', 'id');

    make.senum('mediaType', 'media_type').strmap({
      "@uncaught": IGMediaType.Uncaught.index,
      "1": IGMediaType.Photo.index,
      "2": IGMediaType.Video.index,
      "8": IGMediaType.Carousel.index
    });

    make
        .model('images', 'image_versions2')
        .cls<IGMediaResource>()
        .infilter(['candidates']).list;
    make.model('videos', 'video_versions').cls<IGMediaResource>().list;
    make
        .size('originalSize', '@this')
        .width('original_width')
        .height('original_height');
    make.bit('audible', 'has_audio').unkas(false);
    make.num('duration', 'video_duration').fnum;
    make.model('usertags', 'usertags').cls<IGUsertag>().infilter(['in']).list;
  }).verify((make) {
    make.retnil.nkey("mediaId").throws;
    
    make.str("mediaId");
    make.num("mediaType").inum;
    make.model("images").cls<IGMediaResource>(VERIFY_MODEL_JSON_TAG).list;
    make.model("videos").cls<IGMediaResource>(VERIFY_MODEL_JSON_TAG).list;
    make.size("originalSize");
    make.bit("audible").unkas(false);
    make.num("duration").fnum;
    make.model("usertags").cls<IGUsertag>(VERIFY_MODEL_JSON_TAG).list;
  });

  MM(() => IGMediaCaption()).def((make) {
    make.str('captionId', 'pk');
    make.str('mediaId', 'media_id');
    make.str('userId', 'user_id');
    make.str('text', 'text');
    make.bit('translatable', 'has_translation').unkas(false);
  }).verify((make) {
    make.str("captionId");
    make.str("mediaId");
    make.str("userId");
    make.str("text");
    make.bit("translatable").unkas(false);
  });

  MM(() => IGComment()).def((make) {
    make.retnil.nkey("pk").throws;

    make.str('commentId', 'pk');
    make.str('mediaId', 'media_id');
    make.model('user', 'user').cls<IGUserLite>();
    make.date('createAt', 'create_at').ts10;
    make.str('text', 'text');
    make.bit('liked', 'has_liked_comment').unkas(false);
    make.num('likeCount', 'comment_like_count').inum;
    make.num('childrenCount', 'child_comment_count').inum;
    make
        .model('previewChildren', 'preview_child_comments')
        .cls<IGComment>()
        .list;
    make.model('childrenPagi', '@this').cls<IGPagination>('children_comment');
    make.str('parentCommentId', 'parent_comment_id');
  }).verify((make) {
    make.retnil.nkey("commentId").throws;
    
    make.str("commentId");
    make.str("mediaId");
    make.model("user").cls<IGUserLite>(VERIFY_MODEL_JSON_TAG);
    make.date("createAt").ts10;
    make.str("text");
    make.bit("liked").unkas(false);
    make.num("likeCount").inum;
    make.num("childrenCount").inum;
    make.model("previewChildren").cls<IGComment>(VERIFY_MODEL_JSON_TAG).list;
    make.model("childrenPagi").cls<IGPagination>(VERIFY_MODEL_JSON_TAG);
    make.str("parentCommentId");
  });

  MM(() => IGMedia()).def((make) {
    make.retnil.nkey("id").throws;

    make.str('mediaId', 'id');
    make.senum('mediaType', 'media_type').strmap({
      "@uncaught": IGMediaType.Uncaught.index,
      "1": IGMediaType.Photo.index,
      "2": IGMediaType.Video.index,
      "8": IGMediaType.Carousel.index
    });

    make.bit('story', 'is_reel_media').unkas(false);

    // items
    make
        .model('items', '@this')
        .cls<IGMediaItem>()
        .list
        .skip
        .eqnum('media_type', 8); // media_type(=1|2)
    make
        .model('items', 'carousel_media')
        .cls<IGMediaItem>()
        .list
        .skip
        .neqnum('media_type', 8); // media_type(=8)

    make.model('owner', 'user').cls<IGUserLite>();

    make.bit('savable', 'can_viewer_save');
    make.bit('liked', 'has_liked').unkas(false);
    make.bit('saved', 'has_viewer_saved').unkas(false);
    make.bit('commentsDisabled', 'comments_disabled').unkas(false);

    make.model('caption', 'caption').cls<IGMediaCaption>();

    make.date('takenAt', 'taken_at').ts10;

    make.num('likeCount', 'like_count').inum;
    make.num('commentCount', 'comment_count').inum;
    make.num('viewCount', 'view_count').inum;
    make.num('viewCount', 'viewer_count').inum.skip.nkey("viewer_count");

    make.str('trackingToken', 'organic_tracking_token');

    make.model('topLikers', 'top_likers').cls<IGUserLite>().list;
    make.model('previewComments', 'preview_comments').cls<IGComment>().list;

    make.str('shortCode', 'code');
  }).verify((make) {
    make.retnil.nkey("mediaId").throws;
    
    make.str("mediaId");
    make.num("mediaType").inum;

    make.bit("story").unkas(false);

    // items
    make.model("items").cls<IGMediaItem>(VERIFY_MODEL_JSON_TAG).list;

    make.model("owner").cls<IGUserLite>(VERIFY_MODEL_JSON_TAG);

    make.bit("savable");
    make.bit("liked").unkas(false);
    make.bit("saved").unkas(false);
    make.bit("commentsDisabled").unkas(false);

    make.model("caption").cls<IGMediaCaption>(VERIFY_MODEL_JSON_TAG);

    make.date("takenAt").ts10;

    make.num("likeCount").inum;
    make.num("commentCount").inum;
    make.num("viewCount").inum;

    make.str("trackingToken");

    make.model("topLikers").cls<IGUserLite>(VERIFY_MODEL_JSON_TAG).list;
    make.model("previewComments").cls<IGComment>(VERIFY_MODEL_JSON_TAG).list;

    make.str("shortCode");
  });

  MM(() => IGMediaGrid()).def((make) {
    make.retnil.nkey('layout_content');
    make.push('layout_content');
    make.model('items', 'medias').cls<IGMedia>().list.extract(['media']);
  }).verify((make) {
    make.model("items").cls<IGMedia>(VERIFY_MODEL_JSON_TAG).list;
  });

  MM(() => IGLocation()).def((make) {
    make.retnil.nkey('pk');
    make.str('locationId', 'pk');
    make.str('name', 'name');
    make.str('address', 'address');
    make.str('city', 'city');
    make.str('shortname', 'short_name');
    make.point('coordinate', '@this').x('lng').y('lat');
  }).add("SearchLocation", (make) {
    make.retnil.nkey('location');
    make.push('location');
    make.retnil.nkey('pk');
    make.str('locationId', 'pk');
    make.str('name', 'name');
    make.str('address', 'address');
    make.str('city', 'city');
    make.str('shortname', 'short_name');
    make.point('coordinate', '@this').x('lng').y('lat');
    make.pop;
    make.push('header_media');
    make.model('header', 'media').cls<IGMedia>().listone;
  }).verify((make) {
    make.retnil.nkey('locationId');
    make.str("locationId");
    make.str("name");
    make.str("address");
    make.str("city");
    make.str("shortname");
    make.point("coordinate");
  });

  final iEP = MM.endp;

  iEP.act(IGAPIAction.AuthLogin, (make) {
    make.model('0', 'logged_in_user').cls<IGLoggedInUser>();
  });

  iEP.act(IGAPIAction.AuthFBLogin, (make) {
    make.model('0', 'logged_in_user').cls<IGLoggedInUser>();
  });

  iEP.act(IGAPIAction.AuthTFLogin, (make) {
    make.model('0', 'logged_in_user').cls<IGLoggedInUser>();
  });

  iEP.act(IGAPIAction.AuthTFSmsSend, (make) {
    make.model('0', '@this').cls<IGTwoFactorLock>();
  });

  iEP.act(IGAPIAction.AuthChallengeFetch, (make) {
    make.model('0', '@this').cls<IGChallengeSelection>();
  });

  iEP.act(IGAPIAction.AuthChallengeSelect, (make) {
    make.model('0', '@this').cls<IGChallengeVerification>();
  });

  iEP.act(IGAPIAction.AuthChallengeVerify, (make) {
    make
        .model('0', 'logged_in_user')
        .cls<IGLoggedInUser>()
        .skip
        .nobj("logged_in_user");
  });

  iEP.act(IGAPIAction.AccountProfileFetch, (make) {
    make.model('0', 'user').cls<IGAccountUser>();
  });

  iEP.act(IGAPIAction.AccountPrivacySwitch, (make) {
    make.model("0", "user").cls<IGUserLite>();
  });

  iEP.act(IGAPIAction.AccountCommenterControlFetch, (make) {
    make.senum("0", "audience_control").strmap({
      "@uncaught": IGCommenterControl.Uncaught.index,
      "any": IGCommenterControl.Any.index,
      "following_and_follower": IGCommenterControl.FollowingAndFollower.index,
      "following": IGCommenterControl.Following.index,
      "follower": IGCommenterControl.Follower.index
    }).listone;
  });

  iEP.act(IGAPIAction.AccountCommenterControlSwitch, (make) {

  });

  iEP.act(IGAPIAction.UserGraphFollow, (make) {
    make.retnil.nkey('friendship_status');
    make.model('0', 'friendship_status').cls<IGUserGraphStatus>();
  });

  iEP.act(IGAPIAction.UserGraphUnfollow, (make) {
    make.retnil.nkey('friendship_status');
    make.model('0', 'friendship_status').cls<IGUserGraphStatus>();
  });

  iEP.act(IGAPIAction.UserGraphFetch, (make) {
    make.model('0', '@this').cls<IGUserGraphStatus>();
  });

  iEP.act(IGAPIAction.UserGraphBatchFetch, (make) {
    make.retnil.nkey('friendship_statuses');
    make.model('0', 'friendship_statuses').cls<IGUserGraphStatus>().map;
  });

  iEP.act(IGAPIAction.FollowerFetch, (make) {
    make.retnil.nkey('users');
    make.model('0', 'users').cls<IGUserLite>().list;
    make.model('1', '@this').cls<IGPagination>();
  });

  iEP.act(IGAPIAction.FolloweeFetch, (make) {
    make.retnil.nkey('users');
    make.model('0', 'users').cls<IGUserLite>().list;
    make.model('1', '@this').cls<IGPagination>();
  });

  iEP.act(IGAPIAction.UserInfoFetch, (make) {
    make.retnil.nkey('user');
    make.model('0', 'user').cls<IGUser>();
  });

  iEP.act(IGAPIAction.UserInfoFetchByName, (make) {
    make.retnil.nkey('user');
    make.model('0', 'user').cls<IGUser>();
  });

  iEP.act(IGAPIAction.MediaInfoFetch, (make) {
    make.retnil.nkey('items');
    make.model('0', 'items').cls<IGMedia>().listone;
  });

  iEP.act(IGAPIAction.MediaInfoBatchFetch, (make) {
    make.retnil.nkey('items');
    make.model('0', 'items').cls<IGMedia>().list2map('mediaid');
  });

  iEP.act(IGAPIAction.MediaLikerFetch, (make) {
    make.model('0', 'users').cls<IGUserLite>().list;
  });

  iEP.act(IGAPIAction.MediaCommentFetch, (make) {
    make.model('0', 'comments').cls<IGComment>().list;
    make.model('1', 'caption').cls<IGMediaCaption>();
    make.model('2', '@this').cls<IGPagination>();
  });

  iEP.act(IGAPIAction.MediaLike, (make) {});

  iEP.act(IGAPIAction.MediaUnlike, (make) {});

  iEP.act(IGAPIAction.MediaSave, (make) {});

  iEP.act(IGAPIAction.MediaUnsave, (make) {});

  iEP.act(IGAPIAction.MediaPostedView, (make) {});

  iEP.act(IGAPIAction.MediaStoryView, (make) {});

  iEP.act(IGAPIAction.MediaComment, (make) {
    make.retnil.nkey('comment');
    make.model('0', 'comment').cls<IGComment>();
  });

  iEP.act(IGAPIAction.MediaCommentableSwitch, (make) {});

  iEP.act(IGAPIAction.MediaPostedFeedFetch, (make) {
    make.retnil.nkey('items');
    make.model('0', 'items').cls<IGMedia>().list;
    make.model('1', '@this').cls<IGPagination>();
  });

  iEP.act(IGAPIAction.MediaPostedFeedFetchByName, (make) {
    make.retnil.nkey('items');
    make.model('0', 'items').cls<IGMedia>().list;
    make.model('1', '@this').cls<IGPagination>();
  });

  iEP.act(IGAPIAction.MediaLikedFeedFetch, (make) {
    make.retnil.nkey('items');
    make.model('0', 'items').cls<IGMedia>().list;
    make.model('1', '@this').cls<IGPagination>();
  });

  iEP.act(IGAPIAction.MediaSavedFeedFetch, (make) {
    make.retnil.nkey('items');
    make.model('0', 'items').cls<IGMedia>().list;
    make.model('1', '@this').cls<IGPagination>();
  });

  iEP.act(IGAPIAction.MediaUsertagFeedFetch, (make) {
    make.retnil.nkey('items');
    make.model('0', 'items').cls<IGMedia>().list;
    make.model('1', '@this').cls<IGPagination>();
  });

  iEP.act(IGAPIAction.MediaHashtagFeedFetch, (make) {
    make.retnil.nkey('items');
    make.model('0', 'items').cls<IGMedia>().list;
    make.model('1', '@this').cls<IGPagination>();
  });

  iEP.act(IGAPIAction.MediaStoryFeedFetch, (make) {
    make.retnil.nkey('reel');
    make.model('0', 'reel').cls<IGMedia>().infilter(['items']).list;
  });

  iEP.act(IGAPIAction.MediaLocationFeedFetch, (make) {
    make.retnil.nkey('sections');
    make.model('0', 'sections').cls<IGMediaGrid>().list;
    make.model('1', '@this').cls<IGPagination>('location_feed');
  });

  iEP.act(IGAPIAction.UserSearch, (make) {
    make.model("0", "users").cls<IGUserLite>().list;
    make.bit("1", "has_more").unkas(false);
  });

  iEP.act(IGAPIAction.UploadPhoto, (make) {
    make.retnil.nkey("upload_id").throws;
  });

  iEP.act(IGAPIAction.MediaConfig, (make) {
    make.retnil.nkey('media').throws;
    make.model('0', 'media').cls<IGMedia>();
  });

  iEP.act(IGAPIAction.UploadAvatar, (make) {
    make.retnil.nkey("user").throws;
    make.model("0", "user").cls<IGAccountUser>();
  });
}
