
import 'makers.dart';

class QuarkModel {
  final Map<String, dynamic> _obj = {};

  static dynamic _lower(dynamic obj) {
    if (obj is QuarkModel) {
      return obj.toJson();
    }
    if (obj is List) {
      final list = [];
      obj.forEach((v) {
        list.add(_lower(v));
      });
      return list;
    }
    if (obj is Map) {
      final map = {};
      obj.forEach((k, v) {
        map[k] = _lower(v);
      });
      return obj;
    }
    // null,bool,num,string immutable data
    return obj;
  }

  /// @警告 内部使用
  /// 
  /// 整型转时间类型,均以毫秒存储
  DateTime internalAsDate(String key) {
    dynamic i = _obj[key];
    return DateTime.fromMillisecondsSinceEpoch(i is int ? i : 0);
  }

  /// @警告 内部使用
  /// 
  /// 整型转枚举类型
  T internalAsEnum<T>(String key, List<T> values) {
    dynamic i = _obj[key];
    if (i is! int || i >= values.length) return values[0];
    else return values[i];
  }

  /// @警告 内部使用
  /// 
  /// 绑定内部对象
  internalAsSelf(Map<String, dynamic> obj) {
    _obj.clear();
    _obj.addAll(obj);
    // return this;
  }

  /// 通过 QuarkModel.fromJson<XXX>(obj) 来实例化
  static T? fromJson<T extends QuarkModel>(Map<String, dynamic> obj) {
    return MM.createFromJson(T, obj);
  }

  Map<String, dynamic> toJson() {
    // if (_obj is! Map) return null;
    Map<String, dynamic> json = {};
    _obj.forEach((k, v) {
      json[k] = _lower(v);
    });
    return json;
  }
}

class SXPoint extends QuarkModel {
  num get x => _obj["x"] ?? 0;
  num get y => _obj["y"] ?? 0;
}

class SXSize extends QuarkModel {
  num get width => _obj["width"] ?? 0;
  num get height => _obj["height"] ?? 0;
}

class SXEndpointModel extends QuarkModel {
  dynamic tryGet(int index) => _obj[index.toString()];

  Future<void> tuple0() { return Future.value(); }

  Future<SXTuple1<T0>> tuple1<T0>() => Future.value(SXTuple1<T0>(this));

  Future<SXTuple2<T0, T1>> tuple2<T0, T1>() => Future.value(SXTuple2<T0, T1>(this));

  Future<SXTuple3<T0, T1, T2>> tuple3<T0, T1, T2>() => Future.value(SXTuple3<T0, T1, T2>(this));
}

class SXTuple1 <T0> {
  final T0 item0;

  SXTuple1.value(T0 item) : item0 = item;

  SXTuple1(SXEndpointModel endp) : 
  item0 = endp.tryGet(0);
}

class SXTuple2 <T0, T1> {
  final T0 item0;
  final T1 item1;
  SXTuple2(SXEndpointModel endp) : 
  item0 = endp.tryGet(0),
  item1 = endp.tryGet(1);
}

class SXTuple3 <T0, T1, T2> {
  final T0 item0;
  final T1 item1;
  final T2 item2;
  SXTuple3(SXEndpointModel endp) : 
  item0 = endp.tryGet(0),
  item1 = endp.tryGet(1),
  item2 = endp.tryGet(2);
}

// Instagram

enum IGChallengeStep {
  Uncaught,
  Completed,

  SelectVerifyMethod,
  VerifyByPhone,
  VerifyByEmail,

  DeltaLoginReview, // when choice use IGChallengeChoice.ThisWasMe
}

enum IGChallengeChoice {
  Uncaught,
  Phone,
  Email,
  ThisWasMe
}

enum IGMediaType {
  Uncaught,
  Photo,
  Video,
  Carousel
}

class IGPagination extends QuarkModel {
  bool _first = false;

  String? get maxId => _obj["maxId"];

  dynamic tryGet(String? key) =>  key is String ? _obj[key] : null;

  bool get start => _first;

  bool get turning => maxId != null;

  bool get end => _first != true && maxId == null;

  static IGPagination get home {
    final pagi = IGPagination();
    pagi._first = true;
    return pagi;
  }
}

class IGTwoFactorLock extends QuarkModel {

  /// TFA用户
  String get username => _obj["username"];

  /// TFA识别码
  String get tfaId => _obj["tfaId"];

  /// 手机号后4位
  String get contact => _obj["contact"];
}

class IGChallengeLock extends QuarkModel {

  /// Web版网页路径
  String get url => _obj["url"];

  /// app版API路径
  String get apiPath => _obj["apiPath"];
}

class IGChallengeSelection extends QuarkModel {

  /// 操作步骤
  IGChallengeStep get step => internalAsEnum("step", IGChallengeStep.values);

  /// 认证手机号; 可能为null
  String get phone => _obj["phone"];

  /// 认证邮箱; 可能为null
  String get email => _obj["email"];
}

class IGChallengeVerification extends QuarkModel {

  /// 操作步骤
  /// 在DeltaLoginReview下仅检查该值是否为IGChallengeStep.Completed
  IGChallengeStep get step => internalAsEnum("step", IGChallengeStep.values);

  /// 选择的认证方式
  IGChallengeChoice get choice => internalAsEnum("choice", IGChallengeChoice.values);

  /// 重新发送的时间间隔
  int get resendDelay => _obj["resendDelay"];

  /// 认证方式的模糊标记
  String get contact => _obj["contact"];

  String get userId => _obj["userId"];

  String get nonce => _obj["nonce"];
}

enum IGCommenterControl {
  Uncaught,
  Any,
  FollowingAndFollower,
  Following,
  Follower
}

class IGLoggedInUser extends QuarkModel {
  
  /// 用户ID
  String get userId => _obj["userId"];

  /// 用户名
  String? get username => _obj["username"];

  /// 昵称
  String get nickname => _obj["nickname"];

  /// 头像(缩略图)
  String get thumbnailAvatarURL => _obj["thumbnailAvatarURL"];

  /// 是否使用了匿名头像
  bool get anonymousAvatar => _obj["anonymousAvatar"];

  /// 是否为私有用户
  bool get private => _obj["private"];

  /// 是否为认证用户
  bool get verified => _obj["verified"];

  /// 是否为商业用户
  bool get business => _obj["business"];

  /// 评论者控制
  IGCommenterControl get commenterControl => internalAsEnum("commenterControl", IGCommenterControl.values);

  /// 手机号码, 格式 +{country_code}{national_number}
  String? get phone => _obj["phone"];

  /// facebook绑定的userid
  String? get fbuid => _obj["fbuid"];
}

enum IGUserGender {
  Uncaught,
  NotSpecified,
  Male,
  Female,
}

class IGAccountUser extends QuarkModel {

  /// 用户ID
  String get userId => _obj["userId"];

  /// 用户名
  String get username => _obj["username"];

  /// 昵称
  String get nickname => _obj["nickname"];

  /// 头像(缩略图)
  String get thumbnailAvatarURL => _obj["thumbnailAvatarURL"];

  /// 是否使用了匿名头像
  bool get anonymousAvatar => _obj["anonymousAvatar"];

  /// 是否为私有用户
  bool get private => _obj["private"];

  /// 是否为认证用户
  bool get verified => _obj["verified"];

  /// 评论者控制
  IGCommenterControl get commenterControl => internalAsEnum("commenterControl", IGCommenterControl.values);

  /// 个人简介
  String get biography => _obj["biography"];

  /// 个人网站
  String get website => _obj["website"];

  /// 头像(高清); 150X150 < (320X320; 640X640; ...) < 原图尺寸
  List<IGMediaResource> get narrowedAvatarURLs => _obj["narrowedAvatarURLs"];

  /// 头像(原始)
  IGMediaResource get originalAvatarURL => _obj["originalAvatarURL"];

  /// 手机号码, 格式 +{country_code}{national_number}
  String get phone => _obj["phone"];

  /// 电子邮箱
  String get email => _obj["email"];

  IGUserGender get gender => internalAsEnum("gender", IGUserGender.values);
}

class IGUser extends QuarkModel {

  /// 用户ID
  String get userId => _obj["userId"];

  /// 用户名
  String get username => _obj["username"];

  /// 昵称
  String get nickname => _obj["nickname"];

  /// 头像(缩略图); 150X150
  String? get thumbnailAvatarURL => _obj["thumbnailAvatarURL"];

  /// 是否使用了匿名头像
  bool get anonymousAvatar => _obj["anonymousAvatar"] ?? false;

  /// 是否为私有用户
  bool get private => _obj["private"] ?? false;

  /// 是否为认证用户
  bool get verified => _obj["verified"] ?? false;

  /// 是否为商业用户
  bool get business => _obj["business"] ?? false;

  /// 是否为热门用户(网红?)
  bool get favorite => _obj["favorite"] ?? false;

  /// 发帖数
  int get mediaCount => _obj["mediaCount"] ?? 0;

  /// 被帖子提及的数量
  int get usertagCount => _obj["usertagCount"] ?? 0;

  /// 粉丝数量
  int get followerCount => _obj["followerCount"] ?? 0;

  /// 关注数量
  int get followeeCount => _obj["followeeCount"] ?? 0;

  /// 在instagram拥有关系网?(基本都是true)
  bool get hasChaining => _obj["hasChaining"] ?? false;

  /// 个人简介
  String get biography => _obj["biography"];

  /// 网站
  String get website => _obj["website"];

  /// 头像(高清); 150X150 < (320X320; 640X640; ...) < 原图尺寸
  List<IGMediaResource> get narrowedAvatarURLs => _obj["narrowedAvatarURLs"] ?? [];

  /// 头像(原始)
  IGMediaResource get originalAvatarURL => _obj["originalAvatarURL"];
}

class IGUserLite extends QuarkModel {

  /// 用户ID
  String get userId => _obj["userId"];

  /// 用户名
  String get username => _obj["username"];

  /// 昵称
  String get nickname => _obj["nickname"];

  /// 头像(缩略图)
  String? get thumbnailAvatarURL => _obj["thumbnailAvatarURL"];

  /// 是否使用了匿名头像
  bool get anonymousAvatar => _obj["anonymousAvatar"];

  /// 是否为私有用户
  bool get private => _obj["private"];

  /// 是否为认证用户
  bool get verified => _obj["verified"];
}

class IGMediaResource extends QuarkModel {

  /// URI
  String? get url => _obj["url"];

  /// 尺寸
  SXSize get size => _obj["size"];

  /// video_id ;Video Only
  String get representationId => _obj["representationId"];
}

class IGUserGraphStatus extends QuarkModel {

  /// 我是否正在关注对方
  bool get following => _obj["following"];

  /// 对方是否为私有用户
  bool get private => _obj["private"];

  /// 对方是否向我发出了关注请求
  bool get incoming => _obj["incoming"];

  /// 是否向对方发出了关注请求
  bool get outgoing => _obj["outgoing"];

  /// 对方是否是我的密友
  bool get bestie => _obj["bestie"];

  /// 我是否拉黑了对方; 需要专门接口支持(maybe null);
  bool get blocking => _obj["blocking"];

  /// 对方是否正在关注我; 需要专门接口支持(maybe null)
  bool get followedBy => _obj["followedBy"];
}

class IGUsertag extends QuarkModel {
  
  /// 用户信息
  IGUserLite get user => _obj["user"];

  /// tag的位置
  SXPoint get position => _obj["position"];
}

class IGMediaItem extends QuarkModel {

  /// 帖子ID
  String get mediaId => _obj["mediaId"];

  /// 帖子类型
  IGMediaType get mediaType => internalAsEnum("mediaType", IGMediaType.values);

  /// 图片资源
  List<IGMediaResource> get images => _obj["images"];

  /// 视频资源; Video Only
  List<IGMediaResource> get videos => _obj["videos"];

  /// 原始尺寸
  SXSize get originalSize => _obj["originalSize"];

  /// 是否有音频
  bool get audible => _obj["audible"];

  /// 视频时长; Video Only
  double get duration => _obj["duration"];

  /// 用户标签; Photo Only
  List<IGUsertag> get usertags => _obj["usertags"];
}

class IGMediaCaption extends QuarkModel {

  /// 标题ID
  String get captionId => _obj["captionId"];

  /// 帖子ID
  String get mediaId => _obj["mediaId"];

  /// 发帖人ID
  String get userId => _obj["userId"];

  /// 标题文本
  String get text => _obj["text"];

  /// 是否可翻译
  bool get translatable => _obj["translatable"];
}

class IGComment extends QuarkModel {

  /// 评论ID
  String get commentId => _obj["commentId"];

  /// 评论帖子的ID
  String get mediaId => _obj["mediaId"];

  /// 评论者
  IGUserLite get user => _obj["user"];

  /// 创建时间
  DateTime get createAt => internalAsDate("createAt");

  /// 评论文本内容
  String get text => _obj["text"];

  /// 是否点赞了该评论
  bool get liked => _obj["liked"];

  /// 点赞数
  int get likeCount => _obj["likeCount"];

  /// 评论(回复)数"
  int get childrenCount => _obj["childrenCount"];

  /// 预览回复
  List<IGComment> get previewChildren => _obj["previewChildren"];

  /// 剩余回复页
  IGPagination get childrenPagi => _obj["childrenPagi"];

  /// 回复的评论ID; 如果这是一条回复
  String get parentCommentId => _obj["parentCommentId"];
}

class IGMedia extends QuarkModel {

  /// 帖子ID
  String? get mediaId => _obj["mediaId"];

  /// web专用
  String? get shortCode => _obj["shortCode"];

  /// 帖子类型
  IGMediaType get mediaType => internalAsEnum("mediaType", IGMediaType.values);

  /// 是否为Story=ReelMedia
  bool get story => _obj["story"];

  List<IGMediaItem> get items => _obj["items"];

  /// 发帖人
  IGUserLite get owner => _obj["owner"];

  /// 是否已点赞
  bool get liked => _obj["liked"];

  /// 是否已收藏
  bool get saved => _obj["saved"] ?? false;

  /// viewer是否可以收藏
  bool get savable => _obj["savable"];

  /// 是否关闭了评论
  bool get commentsDisabled => _obj["commentsDisabled"] ?? false;

  /// 帖子的标题
  IGMediaCaption get caption => _obj["caption"];

  /// 发布时间
  DateTime get takenAt => internalAsDate("takenAt");

  /// 点赞数
  int get likeCount => _obj["likeCount"];

  /// 评论数
  int get commentCount => _obj["commentCount"];

  /// 播放数; Video Only
  int get viewCount => _obj["viewCount"];

  /// 与PostView有关的token
  String get trackingToken => _obj["trackingToken"];

  /// 置顶点赞者
  List<IGUserLite> get topLikers => _obj["topLikers"];

  /// 预览评论
  List<IGComment> get previewComments => _obj["previewComments"];

  /// web分享链接
  String? get shareLink => shortCode is String ? "https://www.instagram.com/p/$shortCode/" : null;
}

class IGMediaGrid extends QuarkModel {

  List<IGMedia> get items => _obj["items"];
}

class IGLocation extends QuarkModel {

  /// 地理位置ID
  String get locationId => _obj["locationId"];

  /// 地理名称
  String get name => _obj["name"];

  /// 与city构成subtitle", "address, city"
  String get address => _obj["address"];

  /// 与address构成subtitle
  String get city => _obj["city"];

  String get shortname => _obj["shortname"];

  /// 经纬度坐标, 小数点后6位的精度
  SXPoint get coordinate => _obj["coordinate"];

  /// 关联的facebook地点ID
  String get fbplacesId => _obj["fbplacesId"];

  /// 封面
  IGMedia get header => _obj["header"];

}
